////////////////////////////////////////////////////////////////////////////////
//
//  File          : verify.c
//  Description   : This contains the main function for a verification tool
//                  used to check the output of the CMPSC 311 HW3 SMSA program
//                  by comparing outputs with md5sums. The verification program
//                  uses a known correct file with correct output containing
//                  checksums generated from a correct implementation of the
//                  homework to compare against a student's output file or
//                  piped program output.
//
//   Author : Eric Kilmer
//   Last Modified : Oct 11 20:59:24 EST 2013
//

// Include files
#include <stdio.h>
#include <string.h>

// Project includes


// Defines
// Token which indicates a line to compare
#define OUTPUT_TOK "[OUTPUT]"
#define MAX_LINE_LEN 256

// Output indicators for which line belongs to which file
#define MASTER_IND  "MASTER  >>>>>"
#define STUDENT_IND "STUDENT >>>>>"

#define USAGE \
  "\nUSAGE: [smsasim <workload> 2>&1 | ] verify <master-file> [<student-file>]\n" \
  "Verifies a student's resulting output with the master's output.\n" \
  "Either compare the real-time output of a student's program, or\n" \
  "compare the output files of the student's with the master's.\n" \
  "\n" \
  "where:\n" \
  "   smsasim... - The student's program with options, etc.\n" \
  "   <workload> - The workload file with reads and writes to be tested.\n" \
  "   <master-file> - The correct md5sum values generated by a known correct\n" \
  "                   implementation of smsasim.\n" \
  "   <stduent-file> - The student's output from their program.\n" \
  "\n"

// Fancy color escapes code output
#define ANSI_COLOR_RED     "\x1b[31m"
#define ANSI_COLOR_GREEN   "\x1b[32m"
#define ANSI_COLOR_RESET   "\x1b[0m"

//
// Global Data
char line_master[MAX_LINE_LEN], line_student[MAX_LINE_LEN];


//
// Functional Prototypes
int verify_line( char *master, char *student );
int verify_files( FILE *master, FILE *student );


//
// Functions

////////////////////////////////////////////////////////////////////////////////
//
// Function     : main
// Description  : The main function for the verify program.
//
// Inputs       : argc - the number of command line parameters
//                argv - the parameters
// Outputs      : 0 if successful verification, -1 if mismatch

int main( int argc, char **argv ) {
  FILE *master, *student;
  master = NULL;
  student = NULL;

  unsigned int mismatches;
  
  // Two arguments means that we are comparing two files where the first input
  // is the master and the second one is the student's.
  if ( argc == 3 ) {
    master = fopen( argv[1], "r" );
    student = fopen( argv[2], "r" );

    // Check if we failed to open the files.
    if ( !master ) {
      printf( "Error trying to open master file: [%s]\n", argv[1] );
      return ( -1 );
    }
    if ( !student ) {
      printf( "Error trying to open student file: [%s]\n", argv[2] );
    }
  }
  // 1 argument means that we just have the master file to compare against and
  // we are reading information from the student's program (stdin)
  else if ( argc == 2 ) {
    master = fopen( argv[1], "r" );
    
    // Check if we failed to open the files.
    if ( !master ) {
      printf( "Error trying to open master file: [%s]\n", argv[1] );
      return ( -1 );
    }
  }
  // Incorrect 
  else {
    fprintf( stderr, "Incorrect usage:\n" );
    fprintf( stderr, USAGE );
    return ( -1 );
  }

  printf( "Beginning diff check:\n\n" );
  // Verify that student file matches master file.
  mismatches = verify_files( master, student );

  // Clean up after ourselves.
  fclose( master );
  if ( student && student != stdin ) {
    fclose( student );
  }

  // Return
  if ( mismatches ) {
    return ( -1 );
  }

  return ( 0 );
}


////////////////////////////////////////////////////////////////////////////////
//
// Function     : verify_line
// Description  : Function verifies a single line from the master file and
//                student file.
//
// Inputs       : master - The master's line.
//                student - The student's line.
// Outputs      : 0 if matching, 1 if mismatch 

int verify_line( char *master, char *student ) {
  char *master_out, *student_out;
  int cmp;

  // Get the position of where the signature is found in a line.
  master_out = strstr(master, OUTPUT_TOK);
  student_out = strstr(student, OUTPUT_TOK);

  // If a output token does not exist in the line, then return a mismatch
  if ( !master_out || !student_out ) {
    return ( 1 );
  }

  // Compare the strings to check for a match.
  cmp = strcmp( master_out, student_out );

  // Print out the diff of the master and student
  if ( cmp ) {
    printf( ANSI_COLOR_GREEN "%s %s" ANSI_COLOR_RESET, 
        MASTER_IND, master_out );
    printf( ANSI_COLOR_RED "%s %s\n" ANSI_COLOR_RESET, 
        STUDENT_IND, student_out );
    return ( 1 );
  }

  return ( 0 );
}

////////////////////////////////////////////////////////////////////////////////
//
// Function     : verify_files
// Description  : Verifies the student's file by comparing md5sums against the
//                master's file.
//
// Inputs       : master - the master file.
//                student - the student's file.
// Outputs      : The number of mismatches (0 for success)

int verify_files( FILE *master, FILE *student ) {
  unsigned int mismatched, total;

  // Check if we are reading from an actual student file or from the program
  // output.
  if ( !student ) {
    student = stdin;
  }

  mismatched = 0;
  total = 0;
  // Read through all lines of master and student in compare them while keeping
  // track of both the number of mismatches and total number of comparisons.
  if ( student == stdin ) {
    while ( !feof( master ) ) {
      fgets(line_master, MAX_LINE_LEN, master);
      fgets(line_student, MAX_LINE_LEN, student);
      mismatched += verify_line( line_master, line_student );
      total++;
    }
  }
  else {
    while ( !feof( master ) && !feof( student) ) {
      fgets(line_master, MAX_LINE_LEN, master);
      fgets(line_student, MAX_LINE_LEN, student);
      mismatched += verify_line( line_master, line_student );
      total++;
    }
  }


  // Print out the mismatches and a status of whether the verification was
  // successful or not.
  printf( "Number correct / Total compared: %d/%d\n", total-mismatched, total );
  if ( mismatched ) {
    printf( ANSI_COLOR_RED      "Failed.\n"     ANSI_COLOR_RESET);
  }
  else {
    printf( ANSI_COLOR_GREEN    "Success.\n"    ANSI_COLOR_RESET);
  }

  return ( mismatched );
}
